from joonmyung.meta_data.label import imnet_label, cifar_label
from timm.data import create_dataset, create_loader
from torchvision import transforms
from joonmyung.utils import getDir
import torch
import copy
import glob
import PIL
import os


class JDataset():
    distributions = {"imagenet": {"mean": [0.485, 0.456, 0.406], "std": [0.229, 0.224, 0.225]},
                        "cifar": {"mean": [0.4914, 0.4822, 0.4465], "std": [0.2023, 0.1994, 0.2010]}}
    transform_cifar    = transforms.Compose([transforms.ToTensor(), transforms.Normalize(distributions["cifar"]["mean"], distributions["cifar"]["std"])])
    # transform_imagenet = transforms.Compose([transforms.Resize(256), transforms.CenterCrop(224), transforms.ToTensor(), transforms.Normalize(distributions["imagenet"]["mean"], distributions["imagenet"]["std"])])
    transform_imagenet = transforms.Compose([transforms.ToTensor(), transforms.Normalize(distributions["imagenet"]["mean"], distributions["imagenet"]["std"])])



    transforms = {"imagenet" : transform_imagenet, "cifar" : transform_cifar}

    # CIFAR Setting
    # pip install cifar2png
    # cifar2png cifar100 ./cifar100
    # cifar2png cifar10  ./cifar10
    def validation(self, data):
        return data.lower()

    def unNormalize(self, image):
        result = torch.zeros_like(image)
        for c, (m, s) in enumerate(zip(self.distributions[self.d_kind]["mean"], self.distributions[self.d_kind]["std"])):
            result[:, c] = image[:, c].mul(s).add(m)
        return result

    def normalize(self, image):
        result = copy.deepcopy(image)
        for c, (m, s) in enumerate(zip(self.distributions[self.d_kind]["mean"], self.distributions[self.d_kind]["std"])):
            result[:, c].sub_(m).div_(s)
        return result

    def __init__(self, data_path="/hub_data/joonmyung/data", dataset="imagenet", device="cuda"):
        dataset = dataset.lower()

        self.d      = dataset.lower()
        self.num_classes = 1000 if self.d == "imagenet" else 100
        [self.d_kind, self.d_type] = ["imagenet", "val"] if self.d == "imagenet" else ["cifar", "test"]
        # [self.d_kind, self.d_type] = ["imagenet", "train"] if self.d == "imagenet" else ["cifar", "test"]
        self.device = device

        self.transform = self.transforms[self.d_kind]
        self.data_path = data_path
        self.label_name = imnet_label if self.d_kind == "imagenet" else cifar_label
        self.label_paths = sorted(getDir(os.path.join(self.data_path, self.d_type)))
        self.img_paths = [sorted(glob.glob(os.path.join(self.data_path, self.d_type, label_path, "*"))) for label_path in self.label_paths]

    def __getitem__(self, index=[0,0]):
        label_num, img_num= index
        img_path = self.img_paths[label_num][img_num]
        img = PIL.Image.open(img_path).resize((224, 224))
        data = self.transform(img)

        return data.unsqueeze(0).to(self.device), torch.tensor(label_num).to(self.device), \
                    img, self.label_name[label_num]
    def getItems(self, indexs):
        ds, ls, ies, lns = [], [], [], []
        for index in indexs:
            d, l, i, ln = self.__getitem__(index)
            ds.append(d)
            ls.append(l)
            ies.append(i)
            lns.append(ln)
        return torch.cat(ds, dim=0), torch.stack(ls, dim=0), ies, lns
    def __len__(self):
        return

    def getAllItems(self, batch_size=32):

        dataset = create_dataset(
            root=self.data_path, name="IMNET" if self.d == "imagenet" else self.d.upper()
            , split='validation', is_training=False
            , download=False, load_bytes=False, class_map='')

        loader = create_loader(
            dataset,
            input_size=(3, 224, 224),
            batch_size=batch_size,
            use_prefetcher=True,
            interpolation='bicubic',
            mean=self.distributions[self.d_kind]["mean"],
            std=self.distributions[self.d_kind]["std"],
            num_workers=8,
            crop_pct=0.9,
            pin_memory=False,
            tf_preprocessing=False)
        return loader

if __name__ == "__main__":
    root_path = "/hub_data/joonmyung/data"
    dataset = "cifar100"
    dataset = JDataset(root_path, dataset)
    # sample  = dataset[0, 1]
    samples = dataset.getitems([[0,1], [0,2], [0,3]])





from joonmyung.draw import data2PIL
from joonmyung.utils import to_np
import wandb
import torch
import os
import re

class AverageMeter:
    ''' Computes and stores the average and current value. '''
    def __init__(self) -> None:
        self.reset()

    def reset(self) -> None:
        self.val = 0.0
        self.avg = 0.0
        self.sum = 0.0
        self.count = 0

    def update(self, val: float, n: int = 1) -> None:
        self.val = val
        self.sum += val * n
        self.count += n
        self.avg = self.sum / self.count
    def __str__(self):
        return "\
        end = time.time() \n\
        batch_time = AverageMeter() \n\
        batch_time.update(time.time() - end) \n\
        end = time.time() \n\
        avg_score = AverageMeter()\n\
        accuracy = 0.1\n\
        avg_score.update(accuracy)\n\
        losses = AverageMeter()\n\
        loss = 0\n\
        batch_size = 128\n\
        losses.update(loss.data.item(), batch_size)\n\
        print(f'time {batch_time.val:.3f} ({batch_time.avg:.3f})\t'\n\
              f'loss {losses.val:.4f} ({losses.avg:.4f})\t' \n\
              f'acc {avg_score.val:.4f} ({avg_score.avg:.4f})')"


class Logger():
    loggers = {}
    def __init__(self, use_wandb=True, wandb_entity=None, wandb_project=None, wandb_name=None
                 , wandb_watch=False, main_process=True, wandb_id=None, wandb_dir='./'
                 , args=None, model=False):
        self.use_wandb = use_wandb and main_process

        if self.use_wandb:
            wandb.init(entity=wandb_entity, project=wandb_project, name=wandb_name
                       , save_code=True, resume="allow", id = wandb_id, dir=wandb_dir
                       , config=args, settings=wandb.Settings(code_dir="."))

            if args:
                args.wandb_id = wandb.config.id = wandb.run.id
                torch.save(args, os.path.join(wandb.run.dir, "args.pt"))
            if wandb_watch and model: wandb.watch(model, log='all')


    def getLog(self, k, return_type =None):
        if return_type == "avg":
            return self.loggers[k].avg
        elif return_type == "val":
            return self.loggers[k].val
        else:
            return self.loggers[k]

    def delLog(self, columns: list):
        for column in columns:
            self.loggers.pop(column)

    def resetLog(self):
        self.loggers = {k:AverageMeter() if type(v) == AverageMeter else v for k, v in self.loggers.items()}

    def addLog(self, datas:dict, epoch=None, bs = 1):
        if self.validation():
            for k, v in datas.items():
                data_type = v[0]
                if data_type == 0:  # Values
                    self.loggers[k] = v[1]
                elif data_type == 1: # AverageMeter
                    if k not in self.loggers.keys():
                        self.loggers[k] = AverageMeter()
                    self.loggers[k].update(v[1], bs)
                elif data_type == 2: # Table
                    columns = list(v[1].keys())
                    data_num = len(list(v[1].values())[0])
                    self.loggers[k] = wandb.Table(columns=["epoch"] + columns)
                    for idx in range(data_num):
                        results = []
                        for c in columns:
                            if type(v[1][c]) == torch.Tensor:
                                result = wandb.Image(to_np(data2PIL(v[1][c][idx]))) if len(v[1][c].shape) == 4 else to_np(v[1][c])[idx]
                            elif type(v[1][c]) == list:
                                result = v[1][c][idx]
                            else:
                                raise ValueError
                            results.append(result)
                        self.loggers[k].add_data(str(epoch), *results)
        return True

    def getPath(self):
        return wandb.run.dir

    def logWandb(self):
        if self.validation():
            wandb.log({k: v.avg if type(v) == AverageMeter else v for k, v in self.loggers.items()})
            self.resetLog()
        #
    def finish(self):
        wandb.finish()

    def save(self, file, name):
        if self.validation():
            path = os.path.join(wandb.run.dir, f"{name}.pt")
            torch.save(file, path)
            wandb.save(path, wandb.run.dir)

    def validation(self):
        return True if self.use_wandb else False


if __name__ == "__main__":
    from playground.analysis.lib_import import *
    dataset_name, server, device = "imagenet", "148", "cuda"
    data_path, _ = data2path(server, dataset_name)
    data_num = [[5, 1],
                [5, 2],
                [5, 3],
                [5, 4]]
    dataset = JDataset(data_path, dataset_name, device=device)
    samples, targets, imgs, label_names = dataset.getItems(data_num)
    model = torch.hub.load('facebookresearch/deit:main', "deit_tiny_patch16_224", pretrained=True)

    logger = Logger(use_wandb=True, wandb_entity="joonmyung", wandb_project="test", wandb_name="AAPP",
                    wandb_watch=False, wandb_dir="./")

    logger.addLog({ "sample A": [0, 4],
                    "sample B": [0, 3],
                    "sample C": [0, 2],
                    "table  B": [2, {"image" :     samples, "prediction": targets}]})


    logger.save(model.state_dict(), "checkpoint_best.pt")
    logger.logWandb()
    logger.finish()


from joonmyung.draw import rollout
from joonmyung.meta_data import data2path
from joonmyung.metric import targetPred
from joonmyung.log import AverageMeter
from tqdm import tqdm
from contextlib import suppress
import torch.nn.functional as F
import torch
from joonmyung.utils import to_leaf


class Analysis:
    def __init__(self, model, activate = [True, False, False], detach=True, key_name=None, num_classes = 1000
                 , cls_start=0, cls_end=1, patch_start=1, patch_end=None
                 , ks = 5
                 , amp_autocast=suppress, device="cuda"):
        # Section A. Model
        self.num_classes = num_classes
        self.key_name = key_name
        self.model = model
        self.ks = ks
        self.detach = detach


        # Section B. Attention
        self.kwargs_roll = {"cls_start" : cls_start, "cls_end" : cls_end
                            , "patch_start" : patch_start, "patch_end" : patch_end}


        # Section C. Setting
        hooks = [{"name_i": 'attn_drop', "name_o": 'decoder', "fn_f": self.attn_forward, "fn_b": self.attn_backward},
                 {"name_i": 'qkv', "name_o": 'decoder', "fn_f": self.qkv_forward, "fn_b": self.qkv_backward},
                 {"name_i": 'head', "name_o": 'decoder', "fn_f": self.head_forward, "fn_b": self.head_backward}]
        hooks = [h for h, a in zip(hooks, activate) if a]


        self.amp_autocast = amp_autocast
        self.device = device
        for name, module in self.model.named_modules():
            for hook in hooks:
                if hook["name_i"] in name and hook["name_o"] not in name:
                    module.register_forward_hook(hook["fn_f"])
                    module.register_backward_hook(hook["fn_b"])

    def attn_forward(self, module, input, output):
        # input  : 1 * (8, 3, 197, 197)
        # output : (8, 3, 197, 197)
        self.info["attn"]["f"].append(output.detach() if self.detach else output)

    def attn_backward(self, module, grad_input, grad_output):
        # input  : 1 * (8, 3, 197, 192)
        # output : (8, 3, 197, 576)
        self.info["attn"]["b"].insert(0, grad_input[0].detach() if self.detach else grad_input[0])

    def qkv_forward(self, module, input, output):
        # input  : 1 * (8, 197, 192)
        # output : (8, 197, 576)
        self.info["qkv"]["f"].append(output.detach())

    def qkv_backward(self, module, grad_input, grad_output):
        # self.info["qkv"]["b"].append(grad_input[0].detach())
       pass

    def head_forward(self, module, input, output):
        # input : 1 * (8(B), 192(D)), output : (8(B), 1000(C))
        # TP = targetPred(output.detach(), self.targets.detach(), topk=self.ks)
        TP = targetPred(to_leaf(output), to_leaf(self.targets), topk=self.ks)
        self.info["head"]["TP"] = torch.cat([self.info["head"]["TP"], TP], dim=0) if "TP" in self.info["head"].keys() else TP

    def head_backward(self, module, grad_input, grad_output):
        pass

    def resetInfo(self):
        self.info = {"attn": {"f": [], "b": []}, "qkv": {"f": [], "b": []},
                     "head": {"acc1" : AverageMeter(),
                              "acc5" : AverageMeter(),
                              "pred" : None

                              }}

    def __call__(self, samples, index=None, **kwargs):
        self.resetInfo()
        self.model.zero_grad()
        self.model.eval()

        if type(samples) == torch.Tensor:
            outputs = self.model(samples, **kwargs)
            return outputs
        else:
            for self.inputs, self.targets in tqdm(samples):
                # with self.amp_autocast():
                outputs = self.model(self.inputs)
            return False

    def anaAttn(self, attn=True, grad=False, output=None, index=None,
                head_fusion="mean", discard_ratios=[0.9], data_from="cls"
                , starts=[0], ls=None, bs=None
                , reshape=False, mean=True):

        if attn:
            attn = self.info["attn"]["f"]
        if grad:
            self.info["attn"]["b"] = []
            self.model.zero_grad()
            if index == None: index = output.max(dim=1)[1]
            index = torch.eye(self.num_classes, device=self.device)[index]
            loss = (output * index).sum()
            loss.backward(retain_graph=True)
            grad = self.info["attn"]["b"]

        return rollout(attn, grad
                       , head_fusion=head_fusion, discard_ratios=discard_ratios, data_from=data_from
                       , starts=starts, ls=ls, bs=bs
                       , reshape=reshape, mean=mean)


if __name__ == '__main__':
    # Section A. Data
    dataset_name, server, device, amp_autocast = "imagenet", "148", 'cuda', torch.cuda.amp.autocast
    data_path, _ = data2path(server, dataset_name)
    data_num, batch_size = [[0, 0], [1, 0], [2, 0], [3, 0], [0, 1], [1, 1], [2, 1], [3, 1]], 16
    # data_num, batch_size = [[3, 0]], 16
    # test, activate = [False, False, True], [False, False, True]
    test, activate = [False, True], [True, False, False] # ATTN, QKV, Head

    dataset = JDataset(data_path, dataset_name, device=device)
    samples, targets, imgs, label_names = dataset.getItems(data_num)
    loader = dataset.getAllItems(batch_size)

    # Section B. Model
    # model_number, model_name = 0, "vit_small_patch16_224"
    model_number, model_name = 1, "deit_tiny_patch16_224"
    modelMaker = JModel(model_name, dataset.num_classes, model_number=model_number, device=device)
    model, args = modelMaker.getModel()
    model = Analysis(model, activate = activate, device=device)

    bs = [3]
    samples_ = samples[bs] if bs else samples
    if test[0]:
        ls, starts, col = [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11], None, 12
        discard_ratios, head_fusion, data_from = [0.0, 0.9], "mean", "patch"  # Attention, Gradient
        output = model(samples_)
        rollout_attn = model.anaAttn(True, False, output,
                                     ls=ls, starts=starts, discard_ratios=discard_ratios,
                                     head_fusion = head_fusion, index=targets, data_from=data_from,
                                     reshape=True)

        datas = overlay(samples_, rollout_attn, dataset_name)
        drawImgPlot(datas, col=col)

    if test[1]:
        samples_.requires_grad, model.detach, k = True, False, 3
        output = model(samples_)
        attn = torch.stack(model.info["attn"]["f"], dim=1).mean(dim=[2,3])[0,-2]
        topK = attn[1:].topk(k, -1, True)[1]
        # a = torch.autograd.grad(attn.sum(), samples, retain_graph=True)[0].sum(dim=1)
        a = torch.autograd.grad(output[:,3], samples_, retain_graph=True)[0].sum(dim=1)
        b = F.interpolate(a.unsqueeze(0), scale_factor=0.05, mode='nearest')[0]
        drawHeatmap(b)
        print(1)
        # to_np(torch.stack([attn[:, :, 0], attn[:, :, 1:].sum(dim=-1)], -1)[0])


    if test[2]:
        output = model(loader)


from joonmyung.utils import time2str
from tqdm import tqdm
import subprocess
import time
import pynvml

class GPU_Worker():
    def __init__(self, gpus = [], waitTimeInit = 30, waitTime = 60, count = 0,
                 checkType:int = 0, utilRatio:int = 50, need_gpu=1, reversed=False, p = True):
        self.activate  = False
        self.gpus      = gpus
        self.waitTimeInit = waitTimeInit
        self.waitTime = waitTime
        self.checkType = checkType
        self.need_gpu = int(need_gpu)

        self.reversed  = reversed
        self.utilRatio = utilRatio
        self.p = p
        self.count = count

        self.availGPUs = []

    def getFreeRatio(self, id):
        handle = pynvml.nvmlDeviceGetHandleByIndex(id)
        use = pynvml.nvmlDeviceGetUtilizationRates(handle)
        ratio = 0.5 * (float(use.gpu + float(use.memory)))
        # ratio = float(use.memory)
        # ratio = float(use.gpu)
        return ratio

    def setGPU(self):
        if self.activate: time.sleep(self.waitTimeInit)
        else: self.activate = True

        count = self.count
        pynvml.nvmlInit()
        while True:
            availGPUs = []
            count += 1
            for gpu in self.gpus:
                handle = pynvml.nvmlDeviceGetHandleByIndex(int(gpu))

                # 1. 아무것도 돌지 않는 경우
                if self.checkType == 0 and len(pynvml.nvmlDeviceGetComputeRunningProcesses(handle)) == 0:
                    availGPUs.append(str(gpu))

                # 2. n% 이하를 사용하고 있는 경우
                elif self.checkType == 1 and self.getFreeRatio(int(gpu)) < self.utilRatio:
                    availGPUs.append(str(gpu))

            if len(availGPUs) < self.need_gpu:
                if self.p: print("{} : Wait for finish".format(count))
                time.sleep(self.waitTime)
            else:
                break
        self.availGPUs = availGPUs
        if self.p: print("Activate GPUS : ", self.availGPUs)

    def getGPU(self):
        if len(self.availGPUs) < self.need_gpu: self.setGPU()
        if self.reversed:
            gpus, self.availGPUs = self.availGPUs[:self.need_gpu], self.availGPUs[self.need_gpu:]
        else:
            self.availGPUs, gpus = self.availGPUs[:-self.need_gpu], self.availGPUs[-self.need_gpu:]
        return ','.join(gpus)

def Process_Worker(processes, gpuWorker, p = True):
    start = time.localtime()
    print("------ Start Running!! : {} ------".format(time2str(start)))

    for i, process in enumerate(tqdm(processes)):
        gpu = gpuWorker.getGPU()
        prefix = f"CUDA_VISIBLE_DEVICES={gpu} nohup "
        suffix = f" > {i+1}:gpu{gpu}.log 2>&1 &"
        print("------ {}:GPU{}  {} ------".format(i + 1, gpu, prefix + process + suffix))
        subprocess.call(prefix + process + suffix, shell=True)

    end = time.localtime()
    print("------ End Running!! : {} ------".format(time2str(end)))
    print("Training Time :  : {} ------".format(time2str(end - start)))



if __name__ == '__main__':
    # Wokring Sample
    processes = [1,2,3,4,5]
    gpuWorker = GPU_Worker([0,1,2,3], 30, 120, checkType=1, utilRatio=50, need_gpu=4)
    Process_Worker(processes, gpuWorker)


import torch

def rangeBlock(block, vmin=0, vmax=5):
    loss = torch.arange(vmin, vmax, (vmax - vmin) / block, requires_grad=False).unsqueeze(dim=1)
    return loss

def columnRename(df, ns):
    for n in ns:
        if n[0] in df.columns:
            df.rename(columns = {n[0]: n[1]}, inplace = True)
#     columnRemove(df, ['c1', 'c2' ... ])


def columnRemove(df, ns):
    delList = []
    for n in ns:
        if n in df.columns:
            delList.append(n)
    df.drop(delList, axis=1, inplace=True)
#     columnRename(df, [['c1_p', 'c1_a'] , ['c2_p', 'c2_a']])


def normalization(t, type = 0):
    if type == 0:
        return t / t.max()
    elif type == 1:
        return t / t.min()


from playground.analysis.project.NeurIPS2023.versions.register_models import get_model
from joonmyung.utils import isDir
from timm import create_model
import torch
import os

class JModel():
    def __init__(self, num_classes = None, root_path= None,
                 device="cuda", p=False):
        # Pretrained_Model
        self.num_classes = num_classes

        # Other
        if root_path:
            self.model_path = os.path.join(root_path, "checkpoint_{}.pth")
        if p and root_path:
            print("file list : ", sorted(os.listdir(root_path), reverse=True))
        self.device = device

    def getModel(self, epoch=0, model_number=0, model_name = "deit_tiny", **kwargs):
        model, args = None, None
        if model_number == 0:
            model = create_model(model_name, pretrained=True, num_classes=self.num_classes, in_chans=3, global_pool=None, scriptable=False).to(self.device)
        elif model_number == 1:
            model = torch.hub.load('facebookresearch/deit:main', model_name, pretrained=True).to(self.device)
        elif model_number == 2:
            model_path = self.model_path.format(str(epoch))
            if not isDir(model_path): raise FileExistsError

            checkpoint = torch.load(model_path, map_location='cpu')
            args = checkpoint['args']
            if "token_merging" in kwargs.keys():
                args.task_type[1] = 1 if kwargs["token_merging"] else 0

            model = get_model(args).to(self.device)
            state_dict = []
            for n, p in checkpoint['model'].items():
                if "total_ops" not in n and "total_params" not in n:
                    state_dict.append((n, p))
            state_dict = dict(state_dict)
            model.load_state_dict(state_dict)

        model.eval()

        return model, args





import torch.nn.functional as F
import pandas as pd
import numpy as np
import torch
import os

from joonmyung.meta_data.label import imnet_label, cifar_label


def data2path(server, dataset,
              conference="", wandb_version="", wandb_name=""):
    if "kakao" in server:
        data_path = "/data/opensets/imagenet-pytorch"
        output_dir = "/data/project/rw/joonmyung/conference"
    elif server in ["148", "137", "151", "152", "113", "68", "67", "64"]:
        data_path = "/hub_data1/joonmyung/data"
        output_dir = "/hub_data1/joonmyung/conference"
    elif server in ["154"]:
        data_path = "/data1/joonmyung/data"
        output_dir = "/data1/joonmyung/conference"
    elif server in ["65"]:
        data_path = "/hub_data2/joonmyung/data"
        output_dir = "/hub_data2/joonmyung/conference"
    elif server in ["kisti"]:
        data_path = "/scratch/x2487a05/data"
        output_dir = "/scratch/x2487a05/data"
    else:
        raise ValueError

    if dataset in ["imagenet", "IMNET"]:
        if "kakao" in server:
            data_path = os.path.join(data_path, "imagenet-pytorch")
        else:
            data_path = os.path.join(data_path, "imagenet")


    output_dir = os.path.join(output_dir, conference, wandb_version, wandb_name) if conference else None


    return data_path, output_dir

def get_label(key, d_name ="imagenet"):
    d_name = d_name.lower()
    if d_name in ["imagenet", "IMNET"] :
        return imnet_label[key]
    elif d_name in ["cifar10", "cifar100"]:
        return cifar_label[key]


def makeSample(shape, min=None, max=None, dataType=int, outputType=np, columns=None):
    if dataType == int:
        d = np.random.randint(min, max, size=shape)
    elif dataType == float:
        d = np.random.uniform(low=min, high=max, size=shape)
    else:
        raise ValueError

    if outputType == np:
        return d
    elif outputType == pd:
        return pd.DataFrame(d, columns=None)
    elif outputType == torch:
        return torch.from_numpy(d)

def makeAttn(shape, dim=1):
    return F.softmax(torch.randn(shape), dim=dim)

def set_dtype(df, dtypes):
    for c_n, d_t in dtypes.items():
        if c_n in df.columns:
            df[c_n] = df[c_n].astype(d_t)
    return df
